
/*

====================================================================
                        Copyright Header            
====================================================================

This software module was originally developed by Ralf Funken
(Philips, Eindhoven, The Netherlands) and edited by Takehiro Moriya 
(NTT Labs. Tokyo, Japan) in the course of development of the
MPEG-2/MPEG-4 Conformance ISO/IEC 13818-4, 14496-4. This software
module is an implementation of one or more of the Audio Conformance
tools as specified by MPEG-2/MPEG-4 Conformance. ISO/IEC gives users
of the MPEG-2 AAC/MPEG-4 Audio (ISO/IEC 13818-3,7, 14496-3) free
license to this software module or modifications thereof for use in
hardware or software products. Those intending to use this software
module in hardware or software products are advised that its use may
infringe existing patents. The original developer of this software
module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software
module or modifications thereof in an implementation. Philips retains
full right to use the code for his/her own purpose, assign or donate
the code to a third party. This copyright notice must be included in 
all copies or derivative works. Copyright 2000.


Source file: CONSOLE_IO.C

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "common.h"
#include "console_io.h"

/***************************************************************************/

void print_banner (void)
{
    fprintf (stderr, "\n");
    fprintf (stderr, "ISO/IEC 14496-4 MPEG-4 Conformance Verification\n");
    fprintf (stderr, "%s  version %s\n", PROGRAM_NAME, SSNRCD_VERSION);
    fprintf (stderr, "\n");
}

/***************************************************************************/

static void show_usage (void)
{
    print_banner ();
    fprintf (stderr, "Usage:   %s  [options] <reference inputfile> <inputfile under test>\n", EXECUTABLE_NAME);
    fprintf (stderr, "\n");
    fprintf (stderr, "Valid options are:\n");

    fprintf (stderr, "                -d <num>     Time-offset in samples between reference inputfile and\n");
    fprintf (stderr, "                             inputfile under test. Positive value means inputfile under test\n");
    fprintf (stderr, "                             is delayed over <num> samples and the first <num> \n");
    fprintf (stderr, "                             samples of that input file under test are skipped\n");
    fprintf (stderr, "                             Negative values are permitted.\n");
    fprintf (stderr, "                             Default: %d samples.\n", DEFAULT_DELAY);

    fprintf (stderr, "                -k <num>     Resolution for calculation of RMS and max. abs.\n");
    fprintf (stderr, "                             difference sample, expressed in bits.\n");
    fprintf (stderr, "                             Valid range: [%d...%d]  Default: %d bits.\n", MINIMUM_RESOLUTION, MAXIMUM_RESOLUTION, DEFAULT_RESOLUTION);

    fprintf (stderr, "                -l <num>     Analysis segment length in samples.\n");
    fprintf (stderr, "                             Default: %d samples.\n", DEFAULT_SEGMENT_LENGTH);

    fprintf (stderr, "                -t <num>     Conformance criteria to be applied.\n");
    fprintf (stderr, "                               0 : No criterion (default)\n");
    fprintf (stderr, "                               1 : RMS + Max. absolute difference sample\n");
    fprintf (stderr, "                                   using K bit resolution (see -k option)\n");
    fprintf (stderr, "                               2 : Segmental SNR + avg. Cepstral Distortion\n");
    fprintf (stderr, "                               4 : Number of bits accuracy reached for RMS criterion\n");
    fprintf (stderr, "                             These values may be added. For example -t 7 enables\n");
    fprintf (stderr, "                             all three criteria.\n");

    fprintf (stderr, "                -v           Verbose progress report.\n");

    fprintf (stderr, "\n");
    fprintf (stderr, "If the input files have different length, only the number of samples\n");
    fprintf (stderr, "in the shortest file will be used. The additional samples at the end of\n");
    fprintf (stderr, "the longest file will be ignored.\n");
    fprintf (stderr, "\n");
    fprintf (stderr, "Segmental SNR and cepstral distortion are calculated only for\n");
    fprintf (stderr, "segments with reference signal power in the range [%.0f...%.0f] dB.\n", SSNR_CD_LOWERBOUND, SSNR_CD_UPPERBOUND);
    fprintf (stderr, "Any partial segments at end of input signal are padded with\n");
    fprintf (stderr, "zeroes when calculating segmental SNR and Cepstral Distortion.\n");
    fprintf (stderr, "\n");
}

/***************************************************************************/

void parse_commandline (string           inputfilename1, 
                        string           inputfilename2, 
                        int              *segment_length,
                        int              argc, 
                        char             **argv,
                        unsigned int     *apply_criteria,
                        signed long      *delay,
                        int              *bits_resolution,
                        int              *verbose)

{
   int        i, l, j;
   int        option;
   string     c;
   int        d;
   char       show_usage_flag = FALSE;


   /*  default  */

   (*segment_length)  = DEFAULT_SEGMENT_LENGTH;
   (*apply_criteria)  = DEFAULT_CRITERIA;
   (*delay)           = DEFAULT_DELAY;
   (*bits_resolution) = DEFAULT_RESOLUTION;
   (*verbose)         = FALSE;

   strcpy (inputfilename1, "");
   strcpy (inputfilename2, "");

   /*  n = number of command line arguments  */
   if (argc == 1)
   { 
       show_usage_flag = TRUE;
   }
   else
   { 
       for (i = 1; i < argc; i++)
       {
           strcpy (c, argv[i]);

           if (c[0] == '-')
           {
               l = strlen (c);
               for (j = 1; j < l; j++)
               {
                   option = (char) (c[j]);
                   switch (option)
                   {
                       case 'h': show_usage_flag = TRUE;
                                 break;

                       case 'd': if (i < (argc - 1))
                                 {
                                     (*delay) = atol (argv[++i]);
                                 }
                                 break;    

                       case 'k': if (i < (argc - 1))
                                 {
                                     d = atol (argv[++i]);
                                     if ((d >= MINIMUM_RESOLUTION) && (d <= MAXIMUM_RESOLUTION))
                                     {
                                         (*bits_resolution) = d;
                                     }
                                     else
                                     {
                                         fprintf(stderr,"resolution %d out of range, using default %d\n",d,DEFAULT_RESOLUTION);
                                     }
                                 }
                                 break;

                       case 'l': if (i < (argc - 1))
                                 {
                                     d = atol (argv[++i]);
                                     if (d > 0) 
                                     {
                                         (*segment_length) = d;
                                     }
                                 }
                                 break;    

                       case 't': if (i < (argc - 1))
                                 {
                                     d = atol (argv[++i]);
                                     if (d > 0) 
                                     {
                                         (*apply_criteria) = (unsigned int) d;
                                     }
                                 }
                                 break;

                       case 'v': (*verbose) = TRUE;
                                 break;

                       case '?': show_usage_flag = TRUE;
                                 break;
                   } /* switch */
               } /* for */
           }  /*  if - or / found   */
           else
           {
               if (strcmp (inputfilename1, "") == 0)
               {
                   strcpy (inputfilename1, c);
               }
               else
               {
                   strcpy (inputfilename2, c);
               }
           }
       }
       
       if ((strcmp (inputfilename1, "") == 0) || (strcmp (inputfilename2, "") == 0))
       {
           show_usage_flag = TRUE;
       }
   }

   if (show_usage_flag)
   {
       show_usage ();
       exit (1);
   }
}

/***************************************************************************/
