
/*

====================================================================
                        Copyright Header            
====================================================================

This software module was originally developed by Ralf Funken
(Philips, Eindhoven, The Netherlands) and edited by Takehiro Moriya 
(NTT Labs. Tokyo, Japan) in the course of development of the
MPEG-2/MPEG-4 Conformance ISO/IEC 13818-4, 14496-4. This software
module is an implementation of one or more of the Audio Conformance
tools as specified by MPEG-2/MPEG-4 Conformance. ISO/IEC gives users
of the MPEG-2 AAC/MPEG-4 Audio (ISO/IEC 13818-3,7, 14496-3) free
license to this software module or modifications thereof for use in
hardware or software products. Those intending to use this software
module in hardware or software products are advised that its use may
infringe existing patents. The original developer of this software
module and his/her company, the subsequent editors and their
companies, and ISO/IEC have no liability for use of this software
module or modifications thereof in an implementation. Philips retains
full right to use the code for his/her own purpose, assign or donate
the code to a third party. This copyright notice must be included in 
all copies or derivative works. Copyright 2000.


Source file: SSNRCD.C

*/



/*  Standard includes                           */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/*  Includes for AFsp Library                   */

#include "libtsp.h"
#include "libtsp/AFpar.h"

/*  Private includes                            */

#include "common.h"
#include "console_io.h"
#include "audiofile_io.h"
#include "cepstrum_analysis.h"

/***************************************************************************/

static int compare_audioheaderdata (audioheader header1,
                                    audioheader header2)
{
    int status = 0;

    if (header1.channels != header2.channels) 
    {
        status = 1;
    }

    if (header1.samplefreq != header2.samplefreq)  
    {
        status = 1;
    }

    return (status);
}

/***************************************************************************/

static void compare_audiodata (AFILE               *stream1, 
                               AFILE               *stream2,
                               audioheader         header1,
                               audioheader         header2,
                               unsigned long       samples_to_compare,
                               signed long         delay,
                               int                 segment_length,
                               unsigned long       no_diff_samples [],
                               double              maxabsdif_sample [],
                               double              rms_dif [],
                               double              ssnr [],
                               double              avg_CD [],
                               int                 *max_accuracy,
                               int                 verbose)
{
    int                 sample_counter, n, ch;
    double              abs_difference, abs_difference_scaled;
    double              overall_cum_difference [MAX_PCM_CHANNELS];
    int                 nr_segments_total, segment_counter;
    int                 samples_read1, samples_read2;
    int                 nr_segments_snrcd [MAX_PCM_CHANNELS];
    double              seg_cum_signal1 [MAX_PCM_CHANNELS];
    double              seg_cum_signal2 [MAX_PCM_CHANNELS];
    double              ss_cum [MAX_PCM_CHANNELS];
    float               *pcm_buffer1, *pcm_buffer2, *temp_pcm_buffer;
    double              lpc_coef [N_PR + 1];
    double              cepstrum1 [2*N_PR + 1], cepstrum2 [2*N_PR + 1]; 
    double              cum_sqrt_D [MAX_PCM_CHANNELS];
    double              D;
    double              refsig_power_dB;
    float               *ham_window;
    float               wlag [N_PR + 1];
    int                 skip_length[2];

    /****************************************************************************/
    /*   Memory allocation                                                      */
    /****************************************************************************/
     
    pcm_buffer1     = (float*) malloc (sizeof (float) * header1.channels * segment_length);
    pcm_buffer2     = (float*) malloc (sizeof (float) * header2.channels * segment_length);
    temp_pcm_buffer = (float*) malloc (sizeof (float) * segment_length);
    ham_window      = (float*) malloc (sizeof (float) * segment_length);
   
    if ((pcm_buffer1 == NULL) || 
        (pcm_buffer2 == NULL) || 
        (temp_pcm_buffer == NULL) || 
        (ham_window == NULL))
    {
        printf ("ERROR: Memory allocation failed.\n");
        exit (1);
    }
    

    /****************************************************************************/
    /*   Variable initialization                                                */
    /****************************************************************************/

    for (ch = 0; ch < MAX_PCM_CHANNELS; ch++)
    {
        no_diff_samples[ch]          = 0;
        maxabsdif_sample[ch]         = 0.0;
        nr_segments_snrcd [ch]       = 0;
        rms_dif[ch]                  = INITIAL_RMS_dB;
        overall_cum_difference [ch]  = 0.0;
        seg_cum_signal1 [ch]         = 0.0;
        seg_cum_signal2 [ch]         = 0.0;
        ss_cum [ch]                  = 0.0; 
        ssnr [ch]                    = 0.0;
        cum_sqrt_D [ch]              = 0.0;
        avg_CD [ch]                  = 0.0;
    } 
    
    nr_segments_total = 1 + ((samples_to_compare - 1) / segment_length);
    
    for (n = 0; n <= N_PR; n++)
    {
        lpc_coef [n] = 0.0;
    }

    skip_length[0] = -MIN (0, delay);
    skip_length[1] = MAX (0, delay);

    *max_accuracy = MAXIMUM_RESOLUTION;

    
    /****************************************************************************/
    /*   Calculate windows for cepstral analysis                                */
    /****************************************************************************/

    hamwdw (ham_window, segment_length);
    lagwdw (wlag, N_PR, BW);
   

    /****************************************************************************/
    /*   Process audio input files                                              */
    /****************************************************************************/

    for (segment_counter = 0; segment_counter < nr_segments_total; segment_counter++)
    {

        /****************************************************************************/
        /*   Read one segment of PCM data from both input files                     */
        /****************************************************************************/

        if (verbose)
        {
            fprintf (stderr, "   %d %% \r", (100 * segment_counter) / nr_segments_total);
        }
        
        samples_read1 = get_audiosample (pcm_buffer1, 
                                         stream1, 
                                         header1.channels, 
                                         segment_counter * segment_length + skip_length[0], 
                                         segment_length);

        samples_read2 = get_audiosample (pcm_buffer2, 
                                         stream2,
                                         header1.channels, 
                                         segment_counter * segment_length + skip_length[1], 
                                         segment_length); 

        apply_zeropadding (pcm_buffer1, 
                           header1.channels, 
                           MIN (samples_read1, samples_read2),
                           segment_length);

        apply_zeropadding (pcm_buffer2, 
                           header1.channels, 
                           MIN (samples_read1, samples_read2),
                           segment_length);
        
        /****************************************************************************/
        /*   Calculate max. abs. difference, SNR, RMS, etc. in current segment      */
        /****************************************************************************/

        for (sample_counter = 0; sample_counter < segment_length; sample_counter++)
        {
            for (ch = 0; ch < header1.channels; ch++)
            {
               
                /*   count number of different samples and 
                     calculate max. absolute value in difference signal  */
                
                if (pcm_buffer1[(sample_counter * header1.channels) + ch] == pcm_buffer2[(sample_counter * header1.channels) + ch])
                {
                    abs_difference = 0.0;
                    abs_difference_scaled = 0.0;
                }
                else
                {
                    no_diff_samples [ch]++; 
                    abs_difference = fabs (pcm_buffer1[(sample_counter * header1.channels) + ch] - pcm_buffer2[(sample_counter * header1.channels) + ch]);
                    abs_difference_scaled = abs_difference / PCM_SCALEFACTOR;
                }
                
                if (abs_difference_scaled > maxabsdif_sample [ch])
                {
                    maxabsdif_sample [ch] = abs_difference_scaled;
                }
                
                /*   collect statistics for overall RMS      */
                
                overall_cum_difference [ch] += (abs_difference_scaled * abs_difference_scaled);
                
                /*   collect statistics for segmental SNR    */
                
                seg_cum_signal1 [ch] += (pcm_buffer1[(sample_counter * header1.channels) + ch] / PCM_SCALEFACTOR) * (pcm_buffer1[(sample_counter * header1.channels) + ch] / PCM_SCALEFACTOR);
                seg_cum_signal2 [ch] += (abs_difference_scaled * abs_difference_scaled);
               
            }   /*  for  ch  */
        }
  
        /****************************************************************************/
        /*   Process completed segment                                              */
        /****************************************************************************/
        
        for (ch = 0; ch < header1.channels; ch++)
        {
            if (seg_cum_signal1 [ch] > 0.0)
            {
                refsig_power_dB = 10 * log10 (seg_cum_signal1 [ch] / segment_length);
                
                if ((refsig_power_dB >= SSNR_CD_LOWERBOUND) && (refsig_power_dB <= SSNR_CD_UPPERBOUND))
                {   
                    nr_segments_snrcd[ch]++;

                    ss_cum [ch] += log10 (1.0 + (seg_cum_signal1 [ch] / (SEGSNR_MIN_NOISE + seg_cum_signal2 [ch])));
                    
                    /****************************************************************************/
                    /*   Cepstral analysis on current segment                                   */
                    /****************************************************************************/
                    
                    copy_sample (temp_pcm_buffer, pcm_buffer1, ch, header1.channels, segment_length);
                    calculate_lpc  (temp_pcm_buffer, segment_length, lpc_coef, ham_window, wlag);
                    lpc2cepstrum (N_PR, lpc_coef, 2 * N_PR, cepstrum1);
                    
                    copy_sample (temp_pcm_buffer, pcm_buffer2, ch, header1.channels, segment_length);
                    calculate_lpc  (temp_pcm_buffer, segment_length, lpc_coef, ham_window, wlag);
                    lpc2cepstrum (N_PR, lpc_coef, 2 * N_PR, cepstrum2);
                    
                    for (n = 1, D = 0.0; n <= 2 * N_PR; n++)
                    {
                        D += ((cepstrum1[n] - cepstrum2[n]) * (cepstrum1[n] - cepstrum2[n]));
                    }
                    
                    cum_sqrt_D [ch] += sqrt (D);
                } 
            }
            
            seg_cum_signal1 [ch] = 0.0;
            seg_cum_signal2 [ch] = 0.0;

        }   /*  for  ch  */
    }   /*  for segment_counter   */
    
    
    /****************************************************************************/
    /*   Calculate overall statistics                                           */
    /****************************************************************************/

    /*   calculate overall RMS value of difference signal (in dB)  */

    for (ch = 0; ch < header1.channels; ch++)
    {
        if (overall_cum_difference [ch] > 0.0)
        {
            rms_dif [ch] = 20 * log10 (sqrt (overall_cum_difference [ch] / samples_to_compare));
        }
    }

    
    /*   calculate average segmental SNR value of difference signal (in dB)  */

    for (ch = 0; ch < header1.channels; ch++)
    {
        if ((ss_cum [ch] > 0.0) && (nr_segments_snrcd [ch] > 0))
        {
            ssnr [ch] = 10 * log10 (pow (10.0, (ss_cum [ch] / nr_segments_snrcd [ch])) - 1.0);
        }
        else
        {
            ssnr [ch] = INITIAL_SNR_dB;
        }
    }


    /*   calculate average cepstral distortion (in dB)   */
    
    for (ch = 0; ch < header1.channels; ch++)
    {
        if (nr_segments_snrcd [ch] > 0)
        {
            avg_CD [ch] = ((10.0 * sqrt (2.0)) / (log (10) *  nr_segments_snrcd[ch])) * cum_sqrt_D [ch];
        }
        else
        {
            avg_CD [ch] = 0.0; 
        }
    }


    /****************************************************************************/
    /*   Calculate number of bits accuracy                                      */
    /****************************************************************************/
    
    for (ch = 0; ch < header1.channels; ch++)
    {
        while ((maxabsdif_sample[ch] > ABSDIF_THRESHOLD(*max_accuracy)) && (*max_accuracy > 0))
        {
            (*max_accuracy)--;
        }
        
        if (rms_dif[ch] > THRESHOLD_RMS_dB)
        {
            while ((rms_dif[ch] > RMS_THRESHOLD(*max_accuracy)) && (*max_accuracy > 0))
            {
                (*max_accuracy)--;
            }        
        }
    }
    
    
    /****************************************************************************/
    /*   Free all allocated memory                                              */
    /****************************************************************************/

    free (ham_window);
    free (temp_pcm_buffer);
    free (pcm_buffer2);
    free (pcm_buffer1);
}

/***************************************************************************/

static int  generate_report (int                no_channels,
                             string             input_filename1, 
                             string             input_filename2,
                             unsigned long      samples_to_compare,
                             signed long        delay, 
                             unsigned long      samples_lost[2],
                             int                segment_length,
                             int                bits_resolution,
                             unsigned long      no_diff_samples [],
                             double             maxabsdif_sample [], 
                             double             rms_dif [], 
                             double             ssnr [],
                             double             avg_CD [],
                             int                max_accuracy,
                             unsigned int       apply_criteria)
{

    int                      ch;
    char                     full_acc_fail = FALSE;
    char                     lim_acc_fail = FALSE;
    int                      max_resolution = MAXIMUM_RESOLUTION;
    int                      failure = 0;

    /*  print delay and number of skipped samples    */
    
    if (delay != 0)
    {
        if (delay > 0)
        {   
            printf ("    delay of waveform under test                      : %ld samples\n", delay);
        }                                              
        else
        {
            printf ("    delay of reference waveform                       : %ld samples\n", -delay);
        }
    }
    
    if ((samples_lost[0] > 0) || (samples_lost[1] > 0))
    {
        if (samples_lost[0] > 0)
        {
            printf ("    samples lost at the end of reference waveform     : %ld samples\n", samples_lost[0]);
        }
        else
        {
            printf ("    samples lost at the end of waveform under test    : %ld samples\n", samples_lost[1]);
        }
    }

    /*  print results per channel   */
    
    for (ch = 0; ch < no_channels; ch++)
    {
        printf ("channel: %d \n", ch);

        if (no_diff_samples[ch] == 0)
        {
            printf ("    input channels match exactly\n");
        }
        else
        {
            printf ("    samples compared                              : %ld\n", samples_to_compare);
            printf ("    number of different samples                   : %lu\n", no_diff_samples[ch]);
            
            /*  Max. absolute difference sample             */
            
            printf ("    max. abs. difference sample                   : %.2e (threshold: %.2e, %d bits res.)", maxabsdif_sample[ch], ABSDIF_THRESHOLD(bits_resolution), bits_resolution);
            if ((apply_criteria & RMS_ABSMAXDIFSAMPLE_CRITERION) == RMS_ABSMAXDIFSAMPLE_CRITERION)
            {
                if (maxabsdif_sample[ch] < ABSDIF_THRESHOLD(bits_resolution))
                {
                    printf (" -> OK");
                }
                else
                {
                    printf (" -> FAIL");
                    full_acc_fail = TRUE;
                }
            }
           
            printf ("\n");

            /*  Overall RMS value of difference signal      */

            printf ("    overall RMS value of difference               : ");
            if (rms_dif[ch] > THRESHOLD_RMS_dB)
            {
                printf ("%.2f dB (threshold: %.2f dB, %d bits res.)", rms_dif[ch], RMS_THRESHOLD(bits_resolution), bits_resolution);
                
                if ((apply_criteria & RMS_ABSMAXDIFSAMPLE_CRITERION) == RMS_ABSMAXDIFSAMPLE_CRITERION)
                {
                    if (rms_dif[ch] < RMS_THRESHOLD(bits_resolution))
                    {
                        printf (" -> OK");
                    }
                    else
                    {
                        printf (" -> FAIL");
                        full_acc_fail = TRUE;
                    }
                }
            }
            else
            {
                printf ("--");
            }
            printf ("\n");
            
            /*  Average segmental SNR of difference signal  */
            
            printf ("    average segmental SNR value of difference     : ");
            if (ssnr[ch] > THRESHOLD_SNR_dB)
            {
                printf ("%.2f dB (%d samples per segment)", ssnr[ch], segment_length);
                
                if ((apply_criteria & SEGSNR_CD_CRITERION) == SEGSNR_CD_CRITERION)
                {
                    if (ssnr[ch] > SEGMENTAL_SNR_THRESHOLD)
                    {
                        printf (" -> OK");
                    }
                    else
                    {
                        printf (" -> FAIL");
                        lim_acc_fail = TRUE;
                    }
                }
            }
            else
            {
                printf ("--");
            }
            printf ("\n");

            /*  Average cepstral distortion                 */

            printf ("    cepstral distortion                           : ");
            printf ("%.2f dB (%d samples per segment)", avg_CD [ch], segment_length);
            
            if ((apply_criteria & SEGSNR_CD_CRITERION) == SEGSNR_CD_CRITERION)
            {
                if (avg_CD [ch] < CEPSTRAL_DISTORTION_THRESHOLD)
                {
                    printf (" -> OK");
                }
                else
                {
                    printf (" -> FAIL");
                    lim_acc_fail = TRUE;
                }
            }    
            printf ("\n");
        }
    }


    /*   print overall result   */

    if ((apply_criteria & RMS_ABSMAXDIFSAMPLE_CRITERION) == RMS_ABSMAXDIFSAMPLE_CRITERION)
    {
        printf ("%s  Test on RMS and abs. max. difference sample  : %s\n", GREP_STRING, full_acc_fail ? "FAIL" : "OK");
    }

    if ((apply_criteria & SEGSNR_CD_CRITERION) == SEGSNR_CD_CRITERION)
    {
        printf ("%s  Test on seg. SNR and cepstral distortion     : %s\n", GREP_STRING, lim_acc_fail ? "FAIL" : "OK");
    }

    if ((apply_criteria & NR_BITS_ACCURACY_CRITERION) == NR_BITS_ACCURACY_CRITERION)
    {
        printf ("%s  Reached RMS criteria for a maximum of        : %i bit\n", GREP_STRING, max_accuracy);
    }

    printf ("\n\n");

    if ( full_acc_fail ) failure = 2;
    if ( lim_acc_fail )  failure = 3;

    return ( failure );
}

/***************************************************************************/
/***************************************************************************/

int main (int argc, char *argv[])
{
    string                  input_filename1, input_filename2;
    AFILE                   *stream1;
    AFILE                   *stream2;
    audioheader             audio_header1, audio_header2;

    unsigned long           samples_to_compare;
    signed long             delay;
    unsigned long           start_compare_pos, end_compare_pos;
    unsigned long           samples_lost[2];

    unsigned long           no_diff_samples [MAX_PCM_CHANNELS];
    double                  maxabsdif_sample [MAX_PCM_CHANNELS];
    double                  rms_dif[MAX_PCM_CHANNELS];
    double                  ssnr [MAX_PCM_CHANNELS];
    int                     segment_length;
    double                  avg_CD [MAX_PCM_CHANNELS];
    unsigned int            apply_criteria;
    int                     bits_resolution;
    int                     max_accuracy = 0;
    int                     verbose;
    int                     failure = 0;
  
    /****************************************************************************/
    /*   Get command line parameters                                            */
    /****************************************************************************/

    parse_commandline (input_filename1, 
                       input_filename2, 
                       &segment_length,
                       argc, 
                       argv,
                       &apply_criteria,
                       &delay,
                       &bits_resolution,
                       &verbose);

    /****************************************************************************/
    /*   Program banner on screen                                               */
    /****************************************************************************/

    print_banner ();
    
    printf ("%s  %s and %s:  \n", GREP_STRING, input_filename1, input_filename2);

    /****************************************************************************/
    /*   Open both input files                                                  */
    /****************************************************************************/

    if (open_inputaudio (&stream1, input_filename1, &audio_header1))
    {
        printf ("ERROR: Cannot open reference waveform.\n");
        exit (1);
    }
 
    if (open_inputaudio (&stream2, input_filename2, &audio_header2))
    {
        printf ("ERROR: Cannot open waveform under test.\n");
        exit (1);
    }
        
    /****************************************************************************/
    /*   Check audio header data                                                */
    /****************************************************************************/

    if (compare_audioheaderdata (audio_header1, audio_header2))
    {
        printf ("ERROR: Headers of input files are different.\n");
        exit (1);
    }

    if (check_audioheader (audio_header1))
    {
        printf ("ERROR: Header of reference waveform is invalid.\n");
        exit (1);
    }

    if (check_audioheader (audio_header2))
    {
        printf ("ERROR: Header of waveform under test is invalid.\n");
        exit (1);
    }
    
    if (audio_header1.bits_per_sample != audio_header2.bits_per_sample)
    {    
        printf ("WARNING: Input files have different resolution.\n");
    }

    if ((apply_criteria & RMS_ABSMAXDIFSAMPLE_CRITERION) == RMS_ABSMAXDIFSAMPLE_CRITERION)
    {
        if (bits_resolution >= audio_header1.bits_per_sample)
        {    
            printf ("WARNING: Resolution of reference waveform too low for requested %d bit conformance testing.\n", bits_resolution);
        }

        if (bits_resolution >= audio_header2.bits_per_sample)
        {    
            printf ("WARNING: Resolution of waveform under test too low for requested %d bit conformance testing.\n", bits_resolution);
        }
    }
    
    /****************************************************************************/
    /*   Compare file length                                                    */
    /****************************************************************************/

    start_compare_pos  = (unsigned long) MIN (audio_header1.nrofsamples, MAX (-delay, 0));
    end_compare_pos    = (unsigned long) MAX (MIN (audio_header1.nrofsamples, (audio_header2.nrofsamples - delay)), (long) start_compare_pos);
    samples_to_compare = end_compare_pos - start_compare_pos;

    samples_lost[0] = (unsigned long) MAX ((audio_header1.nrofsamples - end_compare_pos), 0);
    samples_lost[1] = (unsigned long) MAX ((audio_header2.nrofsamples - delay - (signed long) end_compare_pos), 0);

    if (samples_to_compare == 0)
    {  
        printf ("ERROR: Nothing to compare. Either one of the input files has zero length or delay is out of range.\n");
        exit (1);
    }

    if (audio_header1.nrofsamples != audio_header2.nrofsamples)
    {
        printf ("WARNING: Input files have unequal length.\n");
    }


    /****************************************************************************/
    /*   Analyze audio files                                                    */
    /****************************************************************************/

    compare_audiodata (stream1,
                       stream2, 
                       audio_header1, 
                       audio_header2,
                       samples_to_compare, 
                       delay,
                       segment_length,
                       no_diff_samples, 
                       maxabsdif_sample, 
                       rms_dif, 
                       ssnr,
                       avg_CD,
                       &max_accuracy,
                       verbose);


    /****************************************************************************/
    /*   Generate report                                                        */
    /****************************************************************************/

    failure = generate_report (audio_header1.channels,
                               input_filename1, 
                               input_filename2, 
                               samples_to_compare,
                               delay,
                               samples_lost,
                               segment_length,
                               bits_resolution,
                               no_diff_samples, 
                               maxabsdif_sample, 
                               rms_dif, 
                               ssnr,
                               avg_CD,
                               max_accuracy,
                               apply_criteria);

    /****************************************************************************/
    /*   Close both input files                                                 */
    /****************************************************************************/

    close_inputaudio (stream1);
    close_inputaudio (stream2);

    return (failure);
}

/***************************************************************************/
