/**********************************************************************
MPEG-4 Audio VM
Fileformat converter converter

This software module was originally developed by

Olaf Kaehler (Fraunhofer IIS-A)

and edited by


in the course of development of the MPEG-2 AAC/MPEG-4 Audio standard
ISO/IEC 13818-7, 14496-1,2 and 3. This software module is an
implementation of a part of one or more MPEG-2 AAC/MPEG-4 Audio tools
as specified by the MPEG-2 AAC/MPEG-4 Audio standard. ISO/IEC gives
users of the MPEG-2 AAC/MPEG-4 Audio standards free license to this
software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-2 AAC/ MPEG-4 Audio
standards. Those intending to use this software module in hardware or
software products are advised that this use may infringe existing
patents. The original developer of this software module and his/her
company, the subsequent editors and their companies, and ISO/IEC have
no liability for use of this software module or modifications thereof
in an implementation. Copyright is not released for non MPEG-2
AAC/MPEG-4 Audio conforming products. The original developer retains
full right to use the code for his/her own purpose, assign or donate
the code to a third party and to inhibit third party from using the
code for non MPEG-2 AAC/MPEG-4 Audio conforming products. This
copyright notice must be included in all copies or derivative works.

Copyright (c) 2003.
Source file: converter.c



Required modules:
common_m4a.o            common module
cmdline.o               command line module
streamfile.o            common stream file reader/writer

**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "streamfile.h"

#include "common_m4a.h"          /* common module       */
#include "cmdline.h"             /* command line module */

#include "ep_convert.h"
#include "allHandles.h"
#include "flex_mux.h"

#include "mp4au.h"               /* frame work common declarations */


/* maximum number of inputfiles to be processed at once */
#define MAXFILES (8)
#define MAX_LAYER 50
#define MAX_PROGRAMS 8
#define MAX_TRACKS_PER_LAYER 50

static int debugLevel;

static char *outFileName;
static int outFileNameUsed;
static char *oriExt,*bitExt;

static int selectedProgNr, selectedProg;

static char *lowLevelInfo;

static char *epConfigInfo = NULL;
static char *predInfoIn = NULL, *predInfoOut = NULL;
static int removeExtensions;

static int*  varArgIdx;
static CmdLinePara paraList[] = {
  {&varArgIdx,NULL,"<bit stream file(s) (- = stdin)>"},
  {NULL,NULL,NULL}
};

static CmdLineSwitch switchList[] = {
  {"h",NULL,NULL,NULL,NULL,"print help"},
  {"l",&lowLevelInfo,"%s",NULL,NULL,"format specific settings"},
  {"p",&selectedProgNr,"%i",NULL,&selectedProg,"Read only one program"},
  {"o",&outFileName,"%s",NULL,&outFileNameUsed,"output file name (- = stdout)"},
  {"ei",&oriExt,"%s",".mp4",NULL,"input file extension and type"},
  {"eo",&bitExt,"%s",".mp4",NULL,"output file extension and type"},
  {"d",&debugLevel,"%i","0",NULL,"debug level"},
  {"ep",&epConfigInfo,"%s",NULL,NULL,"output epConfig, one value for all layers or individual values"},
  {"predin",&predInfoIn,"%s",NULL,NULL,"predefinition files for conversion to ep2/3"},
  {"predout",&predInfoOut,"%s",NULL,NULL,"write predefinition files while converting from ep2/3"},
  {"rmext",&removeExtensions,NULL,NULL,NULL,"Remove extension payload tracks"},
  {NULL,NULL,NULL,NULL,NULL,NULL}
};


static void stringToVector(char *string, int maxElements, char **vector)
{
  char *fname=string, *tmp;
  int ele;
  for (ele=0; ele<maxElements; ele++) {
    /* go to next word in string */
    while ((fname)&&(*fname==' ')) fname++;
    if (!fname) break;
    /* find end of that word */
    tmp=strchr(fname, ' ');
    /* save in list */
    if (tmp!=NULL) *tmp=0;
    vector[ele] = fname;
    /* continue thereafter */
    if (tmp==NULL) break;
    fname=tmp+1;
  }
}



static HANDLE_STREAMFILE* openInputFiles(char **argv, int *varArgIdx, int *numFiles)
{
  HANDLE_STREAMFILE *files;
  int fileIdx = 0;
  char oriFileName[STRLEN];
  while (varArgIdx[fileIdx] >= 0) fileIdx++;
  files = (HANDLE_STREAMFILE*)malloc(fileIdx*(sizeof(HANDLE_STREAMFILE)));

  fileIdx=0;
  while (varArgIdx[fileIdx] >= 0) {
    /* compose file names */
    if (ComposeFileName(argv[varArgIdx[fileIdx]],0,"",oriExt,oriFileName,STRLEN))
      CommonExit(1,"composed file name too long");
    DebugPrintf(1, "opening %s\n",oriFileName);

    /* open the input files */
    files[fileIdx] = StreamFileOpenRead(oriFileName,FILETYPE_AUTO);
    if (files[fileIdx]==NULL) {
      fileIdx--;
      while (fileIdx>=0) StreamFileClose(files[fileIdx--]);
      free(files);
      return NULL;
    }

    fileIdx++;
  }
  if (numFiles) *numFiles=fileIdx;
  return files;
}

static HANDLE_STREAMPROG* openInputPrograms(HANDLE_STREAMFILE *inFiles,
                                            int fileNum,
                                            int selectProgram,
                                            int *numPrograms)
{
  HANDLE_STREAMPROG *prog;
  int num = 0, j;
  for (j=0; j<fileNum; j++) {
    int progCount = StreamFileGetProgramCount(inFiles[j]);
    if (progCount<0) return NULL;
    DebugPrintf(3, " file has %i programs\n",progCount);
    num+=progCount;
  }
  prog = (HANDLE_STREAMPROG*)malloc(num*(sizeof(HANDLE_STREAMPROG)));

  num=0;
  for (j=0; j<fileNum; j++) {
    int progCount, k;
    /* open all programs */
    progCount = StreamFileGetProgramCount(inFiles[j]);

    for (k=0; k<progCount; k++) {
      if ((selectProgram>=0) && (k != selectProgram)) continue;
      prog[num] = StreamFileGetProgram(inFiles[j], k);
      if (prog[num]==NULL) {
        free(prog);
        return NULL;
      }
      num++;
    }
  }
  if (numPrograms) *numPrograms=num;
  return prog;
}



static HANDLE_EPCONVERTER* initEpConverters(DEC_CONF_DESCRIPTOR* decConf,    /* in */
                                            int numLayer,                    /* in */
                                            int *tracksInLayer,              /* in */
                                            int *extInLayer,                 /* in */
                                            int rmExtTracks,                 /* in */
                                            char *epConfigInfo,              /* in */
                                            char *predInfoIn,                /* in */
                                            char *predInfoOut,               /* in */
                                            DEC_CONF_DESCRIPTOR* decConfOut, /* out */
                                            unsigned long *totalTracksOut)   /* out */
{
  int lay;
  int firstTrackInLayer;
  char *predFile[MAX_LAYER], *writePredFile[MAX_LAYER];
  EP_SPECIFIC_CONFIG epInfoIn[MAX_LAYER],  epInfoOut[MAX_LAYER];
  int                epConfigIn[MAX_LAYER],epConfigOut[MAX_LAYER];
  HANDLE_EPCONVERTER *epConverter = (HANDLE_EPCONVERTER*)malloc(numLayer*sizeof(HANDLE_EPCONVERTER));

  /* get input epConfig values and epSpecificConfigs (and possibly dump them to files) */
  for (lay=0; lay<numLayer; lay++) writePredFile[lay]=NULL;
  if (predInfoOut!=NULL) stringToVector(predInfoOut, numLayer, writePredFile);
  for (firstTrackInLayer=0,lay=0; lay<numLayer; firstTrackInLayer+=tracksInLayer[lay++]) {
    epConfigIn[lay] = decConf[firstTrackInLayer].audioSpecificConfig.epConfig.value;
    epInfoIn[lay]   = decConf[firstTrackInLayer].audioSpecificConfig.epSpecificConfig;
  }

  /* set output epConfig values */
  for (lay=0; lay<numLayer; lay++) predFile[lay]=NULL;
  lay=0;
  if (epConfigInfo!=NULL) {
    stringToVector(epConfigInfo, numLayer, predFile);
    for (; lay<numLayer; lay++) {
      if (predFile[lay]!=NULL) epConfigOut[lay]=atoi(predFile[lay]);
      else break;
    }
    if (lay!=0) {
      int tmp=lay-1;
      for (; lay<numLayer; lay++) epConfigOut[lay]=atoi(predFile[tmp]);
    }
  }
  for (; lay<numLayer; lay++) epConfigOut[lay]=epConfigIn[lay];

  /* read output epSpecificConfigs from pred files */
  for (lay=0; lay<numLayer; lay++) predFile[lay]=NULL;
  if (predInfoIn!=NULL) stringToVector(predInfoIn, numLayer, predFile);

  {
    int j, outTrack;
    for (firstTrackInLayer=0,lay=0,j=0,outTrack=0; lay<numLayer; firstTrackInLayer+=tracksInLayer[lay++]) {
      int k, outTracks;
      char *usePredFile = NULL;
      if (epConfigOut[lay]>=2) {
        if (predFile[j]!=NULL) {
          usePredFile = predFile[j++];
        } else {
          if (epConfigIn[lay]>=2) {
            epInfoOut[lay] = epInfoIn[lay];
          } else {
            CommonWarning("missing pred file for layer %i",lay);
            lay--;
            while (lay>=0) {
              /* TODO: free(*epConfigOut); but not all of them */
              /* TODO: EPconvert_free(epConverter[lay]); does not exist yet */
              lay--;
            }
            free(epConverter);
            return NULL;
          }
        }
      }
      DebugPrintf(5,"layer %i: ep%i -> ep%i\n", lay, epConfigIn[lay], epConfigOut[lay]);
      epConverter[lay] = EPconvert_create(tracksInLayer[lay],
                                          epConfigIn[lay],
                                          &epInfoIn[lay],
                                          writePredFile[lay],
                                          rmExtTracks?extInLayer[lay]:0,
                                          epConfigOut[lay],
                                          &epInfoOut[lay],
                                          usePredFile);
      if (epConverter[lay]==NULL) {
        lay--;
        while (lay>=0) {
          /* TODO: free(*epConfigOut); but not all of them */
          /* TODO: EPconvert_free(epConverter[lay]); does not exist yet */
          lay--;
        }
        free(epConverter);
        return NULL;
      }
      outTracks = EPconvert_expectedOutputClasses(epConverter[lay]);

      for (k=0; k<outTracks; k++) {
        if (outTracks==tracksInLayer[lay]) {
          decConfOut[outTrack] = decConf[firstTrackInLayer+k]; /* memcpy? */
        } else {
          decConfOut[outTrack] = decConf[firstTrackInLayer]; /* memcpy? */
        }
        decConfOut[outTrack].audioSpecificConfig.epConfig.value = epConfigOut[lay];
        decConfOut[outTrack].audioSpecificConfig.epSpecificConfig = epInfoOut[lay];

        /* TODO: mangle bitrates... :( */

        switch (decConfOut[outTrack].audioSpecificConfig.audioDecoderType.value) {
        case AAC_MAIN:
        case AAC_LC:
        case AAC_SSR:
        case AAC_LTP:
        case AAC_SCAL:
        case ER_AAC_LC:
        case ER_AAC_LTP:
        case ER_AAC_LD:
        case ER_AAC_SCAL:
          EPconvert_byteAlignOutput(epConverter[lay]);
          break;
        }

        outTrack++;
      }
    }
    if (totalTracksOut) *totalTracksOut = outTrack;
  }

  return epConverter;
}



/* main -------------*/

int main (int argc, char *argv[])
{
  /* file i/o variables */
  HANDLE_STREAMFILE* inFiles = NULL;
  HANDLE_STREAMPROG* inPrograms = NULL;
  HANDLE_STREAMFILE outFile;
  HANDLE_STREAMPROG* outPrograms = NULL;
  int numPrograms;
  int numFiles;

  /* ep-conversion variables */
  HANDLE_EPCONVERTER *epConverter[MAX_PROGRAMS];
  int numLayer[MAX_PROGRAMS];
  int tracksInLayer[MAX_PROGRAMS][MAX_TRACKS_PER_LAYER];

  /* others */
  int err, prog;

  oriExt = ".mp4";
  bitExt = ".mp4";


  /* ---- evaluate command line ---- */
  {
    char *progName = "<no program name>";

    CmdLineInit(0);
    err = CmdLineEval(argc,argv,paraList,switchList,1,&progName);
    SetDebugLevel(debugLevel);
    if (err) {
      if (err==1) {
        printf("%s\n",progName);
        CmdLineHelp(progName,paraList,switchList,stdout);
        StreamFileShowHelp();
        exit (1);
      } else
        CommonExit(1,"command line error (\"-h\" for help)");
    }

    DebugPrintf(1,"%s\n",progName);
  }

  /* process all files on command line */
  if (varArgIdx[0] < 0) {
    CommonExit(-1,"missing filenames");
  }
  
  /* ---- open input files ---- */
  inFiles = openInputFiles(argv, varArgIdx, &numFiles);
  if (inFiles==NULL) {
    CommonExit(-1,"error opening input");
  }
  inPrograms = openInputPrograms(inFiles, numFiles, (selectedProg?selectedProgNr:-1), &numPrograms);
  if (inPrograms==NULL) {
    CommonExit(-1,"error reading input");
  }

  /* ---- open output file ---- */
  {
    char bitFileName[STRLEN];

    if (outFileNameUsed) {
      if (ComposeFileName(outFileName,0,"",bitExt,bitFileName,STRLEN))
        CommonExit(1,"composed file name too long");
    } else
      if (ComposeFileName(argv[varArgIdx[0]],1,"",bitExt,bitFileName,STRLEN))
        CommonExit(1,"composed file name too long");
    outFile = StreamFileOpenWrite(bitFileName,FILETYPE_AUTO,lowLevelInfo);
    if (outFile == NULL) goto bail_out;
  }

  /* ---- copy and ep-convert program data ---- */
  outPrograms = (HANDLE_STREAMPROG*)malloc(numPrograms*sizeof(HANDLE_STREAMPROG));
  for (prog=0; prog<numPrograms; prog++) {
    int track;
    int extInLayer[MAX_TRACKS_PER_LAYER];
    HANDLE_STREAMPROG prog_data = inPrograms[prog];
    outPrograms[prog] = StreamFileAddProgram(outFile);
    
    {
      /* ---- count layers and tracks ---- */
      int streamCount = prog_data->trackCount;
      int maxLayer = prog_data->trackCount;
      tracksPerLayer(prog_data->decoderConfig, &maxLayer, &streamCount, tracksInLayer[prog], extInLayer);
      numLayer[prog] = maxLayer+1;
    }
    DebugPrintf(3, "input program %i has %i tracks in %i layers ", prog, prog_data->trackCount, numLayer[prog]);

    /* ---- initialise ep-coverter ---- */
    epConverter[prog] = initEpConverters(prog_data->decoderConfig, numLayer[prog],
                                         tracksInLayer[prog], extInLayer, removeExtensions,
                                         epConfigInfo, predInfoIn, predInfoOut,
                                         outPrograms[prog]->decoderConfig, &(outPrograms[prog]->trackCount));
    if (epConverter[prog]==NULL) goto bail_out;

    for (track=0; track<(signed)outPrograms[prog]->trackCount; track++) {
      outPrograms[prog]->dependencies[track] = track-1;
    }
  }

  /* ---- fix/update some header values ---- */
  for (prog = 0; prog < numPrograms; prog++) {
    StreamFileFixProgram(outPrograms[prog]);
  }
      
  /*StreamFileWriteHeader(outFile); (not necessary) */
  {
    /* ---- read and write all access units ---- */
    int frame, layer;
    unsigned long track, firstTrackInLayer, firstTrackOutLayer;
    HANDLE_STREAM_AU inAUs[MAX_TRACKS_PER_LAYER];
    HANDLE_STREAM_AU outAUs[MAX_TRACKS_PER_LAYER];

    for (track=0; track<MAX_TRACKS_PER_LAYER; track++) {
      inAUs[track] = StreamFileAllocateAU(0);
      outAUs[track] = StreamFileAllocateAU(0);
    }
    
    for (frame=0; 1; frame++) {                              /* loop until EOF */
      if (debugLevel==1) fprintf(stderr,"\rframe %i",frame);
      DebugPrintf(2,"frame %ld",frame);
      for (prog = 0; prog < numPrograms; prog++) {           /* loop through all programs in the inputfiles */
        int AUs;
        HANDLE_STREAMPROG inprog = inPrograms[prog];
        DebugPrintf(4," prog  %ld",prog);
        firstTrackInLayer = firstTrackOutLayer = 0;
        for (layer = 0; layer < numLayer[prog]; layer++) {   /* loop through all layers in that program */
          int outTracks = -1;
          DebugPrintf(5," layer  %ld",layer);
          for (AUs = 0; AUs < StreamAUsPerFrame(inprog, firstTrackInLayer); AUs++) {
            outTracks = EPconvert_expectedOutputClasses(epConverter[prog][layer]);
            /* read AUs */
            for (track=0; track<(unsigned)tracksInLayer[prog][layer]; track++) { /* loop through all tracks of that layer */
              DebugPrintf(3," reading track %ld (%ld/%ld)", firstTrackInLayer + track, track, tracksInLayer[prog][layer]);
              err = StreamGetAccessUnit(inprog, firstTrackInLayer + track, inAUs[track]);
              if (err == -2) goto xeof;
              if (err) goto bail;
            }

            /* convert AUs */
            err = EPconvert_processAUs(epConverter[prog][layer],
                                       inAUs, tracksInLayer[prog][layer],
                                       outAUs, MAX_TRACKS_PER_LAYER);
            if (err!=outTracks) {
              if (err<0) goto bail;
              CommonWarning("Expected %i AUs after ep-conversion, but got %i AUs", outTracks, err);
              outTracks=err;
            }

            /* write AUs */
            for (track=0; track<(unsigned)outTracks; track++) { /* loop through all tracks of the output layer */
              DebugPrintf(3," writing track %ld (%ld/%ld)", firstTrackOutLayer + track, track, outTracks);
              err = StreamPutAccessUnit(outPrograms[prog], firstTrackOutLayer + track, outAUs[track]);
              if (err) goto bail;
            }
          }
          firstTrackInLayer += tracksInLayer[prog][layer];
          firstTrackOutLayer += outTracks;
        }
      }
    }

  xeof:

    /* close all files again */
    {
      int fileIdx;
      for (fileIdx=0; fileIdx<numFiles; fileIdx++) {
        StreamFileClose(inFiles[fileIdx]);
      }
      StreamFileClose(outFile);
    }

 bail:
    for (track=0; track<MAX_TRACKS_PER_LAYER; track++) {
      StreamFileFreeAU(inAUs[track]);
      StreamFileFreeAU(outAUs[track]);
    }
    for (prog=0; prog<numPrograms; prog++) {
      for (layer=0; layer < numLayer[prog]; layer++) {
        /* TODO EPconvert_free(epConverter[prog][layer]); */
      }
      free(epConverter[prog]);
    }
  }

 bail_out:
  CmdLineEvalFree(paraList);

  DebugPrintf(1,"finished\n");
  return(0);
}

