#!/usr/local/bin/perl

#
#  MPEG2 AAC Validation Script
#
#
# 
#  Usage:
#     verify.pl  [-decode ] [-verify] [-package]
#
#    -decode : Decodes the Conformance Streams
#    -verify : Compares pcm streams from one directory with streams in 
#              another directory
#    -package: Packages up the decoded pcm streams into compressed archives
#
#  
#  verify.pl -decode
#
#  Using the -decode option, the script will decode all of the MPEG AAC 
#  Conformance Bitstreams using the correct command line parameters.
#
#      - The compressed streams should all be located in one directory, 
#        indicated by $STREAMDIR
#      - The uncompressed streams are all decoded in the $OUTPUTDIR directory.
#
#  All stderr output that results from the decoding process is stored in the
#  $OUTPUTDIR with a .stderr extension.  If there was no significant stderr
#  that file will not exist.
#
#  This script will only decode the compressed file if the output .pcm file
#  does not already exist. In that way, this script can be incrementally run
#  without decoding the same files over and over again.
# 
# verify.pl -verify
# 
# Using the -verify option, the script will perform an RMS comparison on
# all the Sine Wave conformance streams in the $STREAMDIR against similarly 
# names streams in the $VERIFYDIR subdirectory.
#
# If streams do not pass the RMS requirements, a file with an .dif extension
# will be found.
#
# verify.pl -package
#
# This option uses tar and gzip to compress all of the output files located
# in the $OUTPUDIR directory into compressed files and stores them in the
# $ARCHIVEDIR directory.
#
#
#  NOTES for Proper Building of Decoder for L8
#  -------------------------------------------
#  Stream L8 is composed of about 48 channels. In order for the AAC decoder
#  to handle a stream with this many channels, the compiler flag -DBIGDECODER
#  must be set. If this flag is not set, only streams with up to 5.1 channels
#  may be decoded.  The channel definitions are located in chandefs.h
#
# Revision History
#-----------------------------------------------------------------------------
# 980721    Mark Paley     Initial Version
#
#
#
$STREAMDIR = "./streams";
$OUTPUTDIR = "./decoded";
$ARCHIVEDIR = "./archive";
$VERIFYDIR = "./streams/output";
$RMSDIFF = "./rmsdiff";

#command line switches
$fverify = 0; #default no verify
$fdecode = 0; 
$fpackage = 0;

# Global Variabls
$base;  # Basename of a file used in subroutine basename
$cStreamsProcessed = 0; # of Streams processed this run
$cVerifyErrors = 0;


eval ($#ARGV == -1) && die "Usage: verify.pl [-decode] [-package] [-verify]\n";

while ($_ = $ARGV[0],/^-/) { 
    shift;
    /^-ver*/ && $fverify++;
    /^-dec*/ && $fdecode++;
    /^-pac*/ && $fpackage++;

}

#Initialize Directories, etc.
init();

eval ($fdecode == 1) && allfiles("decode");
eval ($fpackage == 1) && allfiles("package");
eval ($fverify == 1) && allfiles("verify");

##  ****** END OF PROGRAM *******



sub allfiles { 

    $mode = $_[0];



#
# SIN1 was archived as s_*, not SIN1_*, so
# it is a special case
#
    print "Processing SIN Streams\n";
while ($filename = <$STREAMDIR/s_*.aac>) { 

    process($filename,"dec -i0",$mode);

}

#
# This loop handles SIN2 and SIN3
#
while ($filename = <$STREAMDIR/SIN?*.aac>) { 

    process($filename,"dec -i0",$mode);

}


    if ($pmode ne "verify") { 
print "Processing Main Profile Streams\n";
while ($filename = <$STREAMDIR/M[12]*.aac>) { 

    process($filename,"dec -i13",$mode);

}

while ($filename = <$STREAMDIR/M3*.aac>) { 
    process($filename,"dec -i0",$mode,"_pce0");
    process($filename,"dec -i1",$mode,"_pce1");

}

print "Processing Low Complexity Streams\n";
while ($filename = <$STREAMDIR/L[1-8]*.aac>) { 

    process($filename,"dec -p LC -i0",$mode);
}


print "Processing SSR Streams, All Bands\n";
while ($filename = <$STREAMDIR/S[1-9]*.aac>) { 

    # Exclude the S18 stream because it is an ADTS STream
    if ($filename=~/S18/){  next;}
#    if ($filename=~/S15/){  next;}
    process($filename,"dec -p SSR -i0",$mode);
}


print "Processing DRC Streams\n";
while ($filename = <$STREAMDIR/drc*.aac>) { 

    process($filename,"dec -d 1 0 -s48000",$mode,"_1");
    process($filename,"dec -d 1 0.707 -s48000",$mode,"_2");
    process($filename,"dec -d 1 1 -s48000",$mode,"_3");
    process($filename,"dec -d .707 0 -s48000",$mode,"_4");
    process($filename,"dec -d .707 .707 -s48000",$mode,"_5");
    process($filename,"dec -d .707 1 -s48000",$mode,"_6");
    process($filename,"dec -d 0 0 -s48000",$mode,"_7");
    process($filename,"dec -d 0 .707 -s48000",$mode,"_8");
    process($filename,"dec -d 0 1 -s48000",$mode,"_9");

}




print "Processing ADTS Streams\n";
while ($filename = <$STREAMDIR/L9_*.aac>) { 

    process($filename,"dec -p SSR -t",$mode);
}

while ($filename = <$STREAMDIR/S18_*.aac>) { 

    process($filename,"dec -p SSR -t",$mode);
}

}

#
# If command line switch indicates package, convert the
# tar files in to tar.gz files
#

if ($mode eq "package") { 
    print "gzipping any tar files in $ARCHIVEDIR\n";
    while (<$ARCHIVEDIR/*.tar>) { 

	system(("gzip $_")) if (-M $_ < -M "$_.gz");

    } 
}


#
#  display Error count if verify
#
    if ($mode eq "verify") { 
	print "Number of Verified files with Errors: $cVerifyErrors\n";
    }

    print "$cStreamsProcessed Streams Processed\n";
    print "The End.\n";

}
############################################################################

sub process { 
    
    local($filename) = $_[0];
    local($cline) = $_[1];
    local($gzarch);
    $found = 0;
    $pmode = $_[2];
    $extra_descrip = $_[3];

    basename($filename);

    # 
    #  Decode The bitstreams
    #
    if ($pmode eq "decode") { 
	$outputfile = $OUTPUTDIR . "/$base" . "$extra_descrip";
	#print "outputfile = $outputfile\n";
    
	if (output_exists($outputfile) == 0 ) { 
	    $info = $outputfile . ".stderr";
	    print "     decoding $filename $extra_descrip\n";    
	    system ( ("$cline $filename $outputfile 2>$info \n"));
	    $cStreamsProcessed++;
	}
	if ( (-s $info) < 30 ) { 
	    unlink($info);
	}
    }

    #
    #  Package the streams into tar.gz files
    #
    if ($pmode eq "package") { 
	$base=~/(SIN\d?|[SLM]\d?\d?|s)/ && ($tar = $1);
	$tarfile = "$ARCHIVEDIR/$tar.tar";
	$gzarch = $tarfile . ".gz";
	# Don't do any tarring if the .gz tar file exists
	next if (-e $gzarch);
	while (<$OUTPUTDIR/$base*.pcm>) { 
	    print "tar uf $tarfile $_\n";
	    system( ("tar uf $tarfile $_"));
	    $cStreamsProcessed++;
	}

    }

    # 
    #  Verify the Streams
    #
    if ($pmode eq "verify") { 
	while (<$OUTPUTDIR/$base*.pcm>) { 
	    /(SIN.*\d?.*|[SLM]\d?\d?.*|s.*)/ && ($target = $1);
	    $target = "$VERIFYDIR/$target";
	    print "Verifying $_ and $target\n";
	    $rval = system( ("rmsdiff $_ $target > $target.dif" ));
	    $diffile = "$target.dif";
	    if (-z $diffile) { 
		unlink($diffile);
	    }
	    else { 
		$cVerifyErrors++;
	    }
	    $cStreamsProcessed++;
	}
    }

 }


sub output_exists { 
    local($filename) = $_[0];
    local($found) = 0;
    while (<$outputfile*.pcm>) { 
	$found = 1;
    }
    $found;
} 

sub basename { 

    local($fname) = $_[0];

    $lpos = 0;
    $pos = $[;

    # find where the last "/" is, so that the beginning of the
    # base of the filename can be extracted   
    while (($pos = index($filename,"\/",$pos)) >= $[) { 
	$lpos = $pos;
	$pos++;
    }

    $fname = substr($fname,$lpos+1);

    $_ = $fname;
    # Remove any extension, if it exists
    if (/(\w*)/) { 
	$fname = $1;
    }
    $base = $fname;
} 


sub init { 


    if ( !$fpackage && !$fverify && !$fdecode) { 
	print "No Valid Options Specified.\n";
	usage();
    }

# First ensure that there is an output directory that can be used to
#store decoded output files.

    mkdir ("$OUTPUTDIR",0777) unless (-d $OUTPUTDIR);
    mkdir ("$ARCHIVEDIR",0777) unless (-d $ARCHIVE);


    if ($fverify) { 
	eval(-d $VERIFYDIR) ||
	die("$VERIFYDIR Not Found. Please Set \$VERIFYDIR variable in script\n".
	    "to point to directory where verificaton streams are located.\n");

	eval (-e $RMSDIFF) || 
	    die ("$RMSDIFF Not Found. rmsdiff source is included with this release.\n");
    }

    eval(-d $STREAMDIR)  ||  
	die("$STREAMDIR Not Found. Please Set \$STREAMDIR variable in script\n".
	    "to point to directory where conformance streams are located.\n");



}





sub usage { 
 
    print "Usage: verify.pl [-verify] [-decode] [-package]\n";
    exit;
}
